/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitaet Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007
*/

#include <ScatterWeb.System.h>
#include <ScatterWeb.CC1020.h>
#include <string.h>

uint16_t interval = 0;			///< Define in witch interval resend packets
uint16_t packet_no = 1;			///< Number of packets
uint16_t Promiscous = 0;		///< Toggle Promiscous mode on/off
uint16_t req_counter = 0;		///< Count the amount of Request Packets
uint16_t rec_counter = 0;		///< Count the amount of received Data Packets
uint16_t DEBUG = 0;				///< Debug mode on/off.

/* ---------------------------------------- VARIABLES (DSR) --------------------------------------------- */

uint16_t list_id[20];			///< List of reached packet id's.
uint16_t list_rep[20];			///< List of reched Route Reply Packets.
uint16_t list_cache_uid[20];	///< List of cached uid.
uint16_t list_cache_data[20];	///< List of cached data.
uint16_t list_cache_dst[20];	///< List of destinations for the data cache.
uint16_t list_cache_try[20];	///< List of tries to send packet.
uint16_t to = 10;				///< DSR Broadcast Destination Adress.
uint16_t seq_no = 1;			///< Uid for Route Request packet.
uint16_t seq_counter = 0;		///< Position of next id in list of reached packet's.
uint16_t rep_counter = 0;		///< Next Position of first reached Route Reply Packet from uid.
uint16_t param;					///< Save destination for the traffic generator Command (RouteTest).
uint16_t oldParam = 0;			///< Last destination for ...

/* ---------------------------------------- TCP --------------------------------------------------------- */

uint16_t max_packets = 100;		///< Set the maximal allowed amount of packets
uint16_t tcp_seq_no = 0;		///< TCP sequenznumber
uint16_t tcp_ack_no = 0;		///< TCP Ack No
uint16_t arg_packet_no = 1;		///< Number of datapackets to transmit
bool server = true;				///< Is node server?
uint16_t list_pos = 0;			///< Save position of next free field on server for cached tcp information.
uint16_t list_pos_c = 0;		///< save position of next free field on client to store sending data.
uint16_t list_payload_s[100];	///< Save TCP payload in list. (Server)
uint16_t list_seq_no_s[100];	///< Save TCP sequence number in list. (Server)
uint16_t list_payload_c[100];	///< Save TCP payload in list. (Client)
uint16_t list_seq_no_c[100];	///< Save TCP sequence number in list. (Client)
uint16_t tcp_adr;				///< Save destination adress.
uint16_t last_ack = 0;			///< Last acknowledged packet.
uint16_t c_window = 1;			///< Current congestion window size.
uint16_t tcp_cached_payload;	///< Cached payload to resend after timeout.
int c_window_counter = 1;		///< Count amount of recieved data packets on server.
bool running_timer = false;		///< Timmer runnig to check fill cw.
bool fin_send = false;			///< FIN packet send/received discard all later received packets.
uint16_t received_ack = 0;		///< Temporaly save the last received ack packet.
uint16_t counter_received = 0;	///< Count the received packets.
uint16_t counter_send = 0;		///< Count the send packets.
uint16_t counter_resend = 0;	///< Count the resend packets.
uint16_t counter_data_t = 0;	///< Count the timeout packets
uint16_t counter_typed = 0;		///< Save typed amount of packets.
uint16_t bigpacket = 0;			///< Send huge packets.
uint16_t big_payload[300];		///< Payload for a big packet.
uint16_t first_lost = 0;		///< Save sequence number of first lost packet.
bool trigger = false;			///< packet loss detected flag
bool established = false;		///< Connection established

/* ---------------------------------------- DATA STRUCTURE (DSR) ---------------------------------------- */

/** Route
*  @brief Datastructure Route Information Packet
*/
struct Route {
	uint16_t source;		///< Source adress.
	uint16_t dest;			///< Destination adress.
	uint16_t id;			///< ID of sender.
	uint16_t hops[11];		///< Sequence of hops.
} route;

/** Data
* @brief Datastructure: Data Packet
*/
struct Data {
	uint16_t* data;			///< Payload.
	uint16_t hops[11];		///< Sequence of hops.
} data;

/* ---------------------------------------- DATA STRUCTURE  (DSR) ------------------------------------------------ */

/* ---------------------------------------- DATA STRUCTURE  (TCP) ------------------------------------------------ */
/** TCP
* @brief Datastructure: TCP
*/
struct TCP {
	uint16_t seq_no;		///< Sequence Number
	uint16_t ack_no;		///< Acknowledgement Number
	bool ack;				///< ACK Flag
	bool fin;				///< FIN Flag
	bool sync;				///< SYNC Flag
	uint16_t payload;		///< Data in TCP Packet
	uint16_t cw;			///< Congestion Window
} tcp_in, tcp_out;

/* ---------------------------------------- DATA STRUCTURE  (TCP) ------------------------------------------------ */

/* ---------------------------------------- METHODS (DSR + TCP) ------------------------------------------------ */

/**
 * @brief Check
 * Check if the destination node is out of range
 */
 
void check(uint16_t val_3)
{
	if(established == true)
	{
		if(server == false && val_3 != last_ack)
		{
			val_3 = last_ack;
			Timers_add(30*1024, check, val_3);
		}
		else if(server == true && val_3 != tcp_out.ack_no)
		{
			val_3 = tcp_out.ack_no;
			Timers_add(30*1024, check, val_3);
		}
		else if(val_3 == tcp_out.ack_no || val_3 == last_ack) 
		{
			printf("Connection to %u failed !\r\n", tcp_adr);
			fin_send_reset();
		}			
	}
}

/**
* @brief Resend
* Resend packets if no acknowledgement received earlier.
* @param val_t sequence number of the packets
*/
void Resend(uint16_t val_t)
{
	if(fin_send == false && established == true)
	{
		if(DEBUG == 1)
		{
			printf("TIMER (Resend)");
			printf("tcp_out.seq_no: %u tcp_seq_no: %u\r\n", tcp_out.seq_no, tcp_seq_no);
		}
		int i;
		for(i = 0; i < max_packets; i++)
		{
			if(list_seq_no_c[i] == val_t)
			{
				tcp_out.seq_no = val_t;
				tcp_out.ack_no = tcp_ack_no;
				tcp_out.sync = false;
				tcp_out.ack = true;
				tcp_out.fin = false;
	
				tcp_seq_no = tcp_out.seq_no;
				tcp_ack_no = tcp_out.ack_no;
	
				// Load correct payload.
				tcp_cached_payload = list_payload_c[i];
	
				if(DEBUG == 1)
				{
					printf("Resend Client TCP\r\n");
					printf("seq_no: %u\r\nack_no: %u\r\nsync: %u\r\nack: %u\r\nfin: %u\r\n", tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
				}
	
				// Resend data.
				ModifyTCP(tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
				// Count resend packet.
				counter_resend++;
				// Start new Timeout-Timer.
				Timers_add(10*1024, Resend, val_t);
			}
		}
	}
}

/**
* @brief data_timeout 
* Timer on server side to send an ack after five seconds.
* 
* @param size Size of the current congestion window.
*/
void data_timeout(int size)
{
	if(fin_send == false && established == true)
	{

		// Only one timer.
		running_timer = true;
	
		if(c_window_counter != size)
		{
			if(DEBUG == 1)
			{
				printf("Data_Timeout: c_win_c: %u size: %u\r\n", c_window_counter, size);
			}

			Timers_add(10*1024, data_timeout, c_window_counter);
		}
		else
		{	
			// Check if we lost a packet on current transmission.
			int e;
			trigger = false;
			first_lost = 0;
			for(e = 0; e < max_packets; e++)
			{
				if(list_seq_no_s[e] == 0 && trigger == false)
				{
						trigger = true;
						first_lost = (list_seq_no_s[(e-1)])+1;
						
				}
			}
			// Remove timer for packets that are allready acknowledged.
			int j;
			for(j = c_window_counter; j > c_window - c_window_counter; j--)
			{
				if(DEBUG == 1)
				{
					printf("Remove Timer %u. (timeout)\r\n", j);
					printf("First lost %u\r\n", first_lost);
				}
				Timers_remove(Resend, j);
			}
			
			// Send ack for the first lost packet with ack number of the last received data packet.
			if(trigger == true)
			{
				tcp_out.ack_no = first_lost;
			}
			else
			{
				tcp_out.ack_no = ++tcp_in.seq_no;
			}
			
			tcp_out.seq_no = ++tcp_seq_no;
			
			if(DEBUG == 1)
			{					
				printf("Timeout triggerd, send ACK packet.\r\n");
				printf("tcp_out.ack_no: %u\r\n" ,tcp_out.ack_no);
				printf("tcp_out.seq_no: %u\r\n" ,tcp_out.seq_no);
				printf("c_window: %u\r\n", c_window);
			}
	
			tcp_out.sync = false;
			tcp_out.ack = true;
			tcp_out.fin = false;
		
			c_window = 1;
			tcp_out.cw = 1;
			c_window_counter = 1;
	
			ModifyTCP(tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
			// Count data timeouts.
			counter_data_t++;
			// Reset congestion window counter.
			running_timer = false;

			// Resend ack packet
			Timers_add(10*1024, data_timeout, size);
		}
	}
}

/**
 * @brief Timer to reset fin_send after 30 ticks. 
 */
 void fin_send_reset()
 {
	if(DEBUG == 1)
	{
		printf("RESET FIN STATE\r\n");
	}
	// Reset all counters
 	fin_send = false;
	received_ack = 0;		
	counter_received = 0;		
	counter_send = 0;		
	counter_resend	= 0;		
	counter_data_t	= 0;
	established = false;
 }

/**
 * @brief Handle incoming radio data.
 */
bool Process_radioHandler(struct netpacket_handler_args* args)
{
	/*
		Notes: * The received packet data becomes invalid when this handler is completed.
	*/

	// Pass packet through command interface
	if ( Cmd_radioHandler( args ) ) {
		// packet handled by messaging module
	}

	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);

	// Packet reached in our id range. ( 1 - 10 )
	if ( args->cpaResult == ADRCOMP_OTHER && args->netheader->from < 11 && args->netheader->to < 11) {
		
		// Packet is Route Request Packet.
		if(args->netheader->type == REQUEST_PACKET)
		{
			// Copy Route Request Packet.
			memcpy(&route,args->payload,args->payload_size);

			// If packet is reached earlier?
			uint16_t i;
			bool set = false;

			uint16_t result;

			// If incoming uid is in local chache ...
			for(i = 0; i < seq_counter+1; i++ )
			{
				if(route.id == list_id[i])
				{
					result = list_id[i];
					set = true;
				}
			}

			// ... discard packet.
			if(set == true)
			{
				if(DEBUG == 1)
					printf(" Discard Route Request packet from source %u.\r\n", route.source);
				return false;
			}

			// Otherwise ...
			// If destination is reached, send Route Reply packet back to source.
			if(route.dest == npsargs.netheader.from)
			{
				bool set = false;

				uint16_t j;
				for(j = 0; j < rep_counter + 1; j++)
				{
					if(list_rep[j] == route.id)
					{
						set = true;
					}
				}

				// Discard duplicated Route Reply Packets.
				if(set == true)
				{
					if(DEBUG == 1)
						printf("Discard Route Reply packet from %u.\r\n", route.dest);
					return false;
				}
				// Send first received Route Reply Packet back to source.
				else
				{
					list_rep[rep_counter = (rep_counter + 1) % 20] = route.id;
					SendRouteReply(args);	
				}
				
			}
			// If destination is not reached, broadcast Route Request Packet.
			else
			{
				SendRouteRequest(route.dest, 1);
			}
		}
	}
	else {
		if ( args->netheader->from < 11 && args->netheader->to < 11) {
		 
				switch ( args->netheader->type ) {

				case REPLY_PACKET:
					// Copy Route Request Packet.
					memcpy(&route,args->payload,args->payload_size);
					SendRouteReply(args);
					break;

				case DATA_PACKET:
					memcpy(&data,args->payload,args->payload_size);
					SendRouteData(args, 0);
					break;

				case PING_REQUEST_PACKET:
					// send back to receiver's address
					Net_sendPong(args->netheader->from);
					break;

				case PING_REPLY_PACKET:
					LOG_LOW( "PONG from %u", args->netheader->from );
					break;

				case TCP_PACKET:
					// Copy TCP Packet.
					memcpy(&tcp_in,args->payload,args->payload_size);
					// Save TCP ack from incoming tcp packet.
					received_ack = tcp_in.ack_no;
					// Count received packet.
					counter_received++;
					// Fall: Verbindungsaufbau
					if(tcp_in.sync)
					{
						// new connection, reset fin
						fin_send = false;
						// Fall: Verbindungsaufbau ( 2. Server -> Client )
						if(DEBUG == 1)
						{
							printf("SYNC\r\n");
						}
						if(tcp_in.ack)
						{
							if(DEBUG == 1)
							{
								printf("Last Ack: %u\r\nAck: %u\r\n", last_ack, tcp_in.ack_no);
							}
							last_ack = tcp_in.ack_no;
							if(DEBUG == 1)
							{
								printf("Last Ack: %u\r\nAck: %u\r\n", last_ack, tcp_in.ack_no);
							}
							
							tcp_out.ack_no = ++tcp_in.seq_no;
							tcp_out.seq_no = ++tcp_seq_no;
							tcp_out.sync = false;
							tcp_out.ack = true;
							tcp_out.fin = false;
							// Initialize congestion window with 1.
							tcp_out.cw = 1;

							if(DEBUG == 1)
							{
								printf("Server -> Client\r\n");
								printf("seq_no: %u\r\nack_no: %u\r\nsync: %u\r\nack: %u\r\nfin: %u\r\n", tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
							}
							// Decrement packets counter.
							arg_packet_no--;
							ModifyTCP(tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
							Timers_add(20*1014, check, last_ack);
							established = true;
							return true;
						}
						// Fall: Verbindungsaufbau ( 1. Client -> Server )
						else
						{
							if(DEBUG == 1)
							{
								printf("1. Client -> Server\r\n");
							}

							tcp_adr = args->netheader->from; // Save client adress for tcp communications.

							if(DEBUG == 0)
							{
								printf("Incoming Transmission from %u\r\n", tcp_adr);
							}
														
							tcp_out.ack_no = ++tcp_in.seq_no;
							tcp_out.seq_no = tcp_seq_no = 100;
							tcp_out.sync = true;
							tcp_out.ack = true;
							tcp_out.fin = false;
							// Initialize congestion window with 1.
							tcp_out.cw = 1;
							
							// Save the incoming ack number.
							tcp_ack_no = tcp_in.ack_no;

							if(DEBUG == 1)
							{
								printf("seq_no: %u\r\nack_no: %u\r\nsync: %u\r\nack: %u\r\nfin: %u\r\n", tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
							}

							ModifyTCP(tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
							established = true;
							Timers_add(20*1014, check, tcp_out.ack_no);
							return true;
						}
					}
					
					// Fall: Daten senden ( 3. Client -> Server )
					if(fin_send == false && established == true)
					{
						if(tcp_in.ack && !tcp_in.sync && !tcp_in.fin)
						{
							// Client
							if(server == false)
							{
								// Calculate new congestion window size.
												
								if(tcp_in.cw < 4)
									c_window = tcp_out.cw = 2 * tcp_in.cw;
								c_window_counter = 1;
								
								if(DEBUG == 1)
								{
									printf("Last Ack: %u\r\nAck: %u\r\n", last_ack, tcp_in.ack_no);
								}
								// Save the last ack number.
								last_ack = received_ack;
								if(DEBUG == 1)
								{
									printf("Last Ack: %u\r\nAck: %u\r\n", last_ack, tcp_in.ack_no);
									printf("3. Client -> Server\r\n");
									printf("server: %u last_ack: %u list_pos_c: %u\r\n", server, last_ack, list_pos_c);
								}
								// It was already acknowledged the first packet. Check cache first time after the primary packet was send. 
								if(last_ack > 0)
								{
									if(DEBUG == 1)
									{
										printf("Delete packets ...\r\n");
										printf("Client:\r\n");
									}
	
									int i;
									for(i = 0; i < max_packets; i++)
									{
										// Delete all packets in cache lower than ack number.
										if(list_seq_no_c[i] < last_ack && last_ack > 0)
										{	
											list_seq_no_c[i] = 0;
											// Remove timer for packets that are allready acknowledged.
											Timers_remove(Resend, i);
										}
										if(DEBUG == 1)
										{
											printf("Cache[%u]=%u\r\n", i, list_seq_no_c[i]);
										}
									}
									// Congestion window is clear, send waiting data.
									TCPData(tcp_seq_no);
									return true;
								}  
							}
	
							// Server registriert ankommende Daten
							else
							{
								if(DEBUG == 1)
								{
									printf("Server: Incoming tcp data ...\r\n");
									printf("seq_no: %u\r\nack_no: %u\r\nsync: %u\r\nack: %u\r\nfin: %u\r\ntcp_out.cw: %u\r\npayload: %u\r\n", tcp_in.seq_no, tcp_in.ack_no, tcp_in.sync, tcp_in.ack, tcp_in.fin, tcp_in.cw, tcp_in.payload);
									printf("Server:\r\n");
								}
								
								// Save congestion window.
								c_window = tcp_in.cw;
	
								// Paket bestaetigen
								bool ok = true;
								
								if(DEBUG == 1)
								{
									printf("Caching:\r\n");
									printf("seq_no-2: %u\r\nseq_no: %u\r\nPayload: %u\r\n", tcp_in.seq_no-2, tcp_in.seq_no, tcp_in.payload);
								}								
								
								list_payload_s[tcp_in.seq_no-2] = tcp_in.payload; 
								list_seq_no_s[tcp_in.seq_no-2] = tcp_in.seq_no;
	
								// Print the current cache content.
								if(DEBUG == 1)
								{
									int i;
									for(i = 0; i < max_packets; i++)
									{
										if(list_seq_no_s[i] != 0)
											printf("Cache[%u]=%u %u\r\n", i, list_seq_no_s[i], list_payload_s[i]);
									}
								}
																
								// Check if we lost a packet on current transmission.
								int e;
								for(e = 0; e < last_ack + c_window; e++)
								{
									if(list_seq_no_s[e] == 0)
										ok = false;
								}
								
								// Only send an ack if we have received all already send packets.
								if(c_window_counter == c_window && ok == true)
								{
									tcp_out.ack_no = ++tcp_in.seq_no;
									tcp_out.seq_no = ++tcp_seq_no;
									
									if(DEBUG == 1)
									{						
										printf("Outgoing tcp data ...\r\n");
										printf("tcp_out.ack_no: %u\r\n " ,tcp_out.ack_no);
										printf("tcp_out.seq_no: %u \r\n" ,tcp_out.seq_no);
										printf("c_window: %u\r\n", c_window);
									}
	
									tcp_out.sync = false;
									tcp_out.ack = true;
									tcp_out.fin = false;
								
									// Remove old timer.
									Timers_remove(data_timeout, c_window);
									// No timer is runnig now.
									running_timer = false;
									// Start Timer to send an ack after a period of time if no other data packets arrieved.
									if(running_timer == false)
	                                {
	                                        if(DEBUG == 1)
	                                        {
	                                                printf("Starting Server Timer!!\r\n");
	                                        }
	                                        data_timeout(0);
	                                }

									ModifyTCP(tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
									// Reset congestion window counter.
									c_window_counter = 1;
									return true;
								}
								//
								else
								{
									if(ok == true)
									{
									// Handle packets that doesn't need an ack packet.
									if(DEBUG == 1)
									{						
										printf("Received tcp data ... (< congestion window)\r\n");
										printf("tcp_out.ack_no: %u\r\n" ,tcp_out.ack_no);
										printf("tcp_out.seq_no: %u\r\n" ,tcp_out.seq_no);
										printf("c_window: %u (%u)\r\n", c_window, c_window_counter);
									}
	
									c_window_counter++;
									tcp_seq_no++;
									return true;
									}
									return false;
								}
							}
						}
		
						// Fall: Verbindung abbauen (Server)
						if(tcp_in.fin)
						{
							if(DEBUG == 1)
							{
								// Empfangene Daten ausgeben.
								printf("IN FIN\r\nTable: ");
								int i=0;
								while(i < max_packets)
								{
									if(list_seq_no_s[i] != 0)
									{
										printf("%u ", list_seq_no_s[i]);
										printf("%u ", list_payload_s[i]);
									}
									i++;
								}
								printf("\r\n");
							}
							
							if(DEBUG == 0)
							{
								printf("Data transmission finished!\r\n");
								printf("Cache: ");
								// Empfangene Daten ausgeben.
								int i=0;
								while(i < max_packets)
								{
									if(list_seq_no_s[i] != 0)
									{
										printf("%u ", list_seq_no_s[i]);										
										printf("%u ", list_payload_s[i]);

									}
									i++;
								}
								printf("\r\n");
								printf("Stats: ");
								printf("received packets: %u, send packets: %u, timeouts occured: %u\r\n", counter_received, counter_send, counter_data_t);
							}							
	
							// Reset nach Verbindungsabbau
							tcp_seq_no = 0;
							tcp_ack_no = 0;
                            last_ack = 0;
                            c_window = 1;
                            c_window_counter = 1;
                            running_timer = false;
                            received_ack = 0;		
							counter_received = 0;		
							counter_send = 0;		
							counter_resend	= 0;		
							counter_data_t	= 0;

                           // Delete cached content
							int i = 0;
							while(i < 51)
							{
								list_payload_s[i] = 0;
								list_seq_no_s[i]  = 0;
								i++;
							}
							
							// List is clear. 
							list_pos = 0;
							// Set fin_send
							fin_send = true;
							// start timer to reset fin state after 30s
							Timers_add(30*1024, fin_send_reset, 0xffee);
							return true;
	
						}
					}
				default:
					return false;		// packet not handled (unknown)
				}
		}
		// process packets for this node here
		
		
		return true;				// packet handled
	}
	return false;					// packet not handled (other destination)
}

/**
 * @brief unused sensorHandler function
 */
void Process_sensorHandler(void* sdata)
{

}

/**
* @brief Method: Validate
*					
* Resend route request packet every 5 seconds 
* if cache is not empty
*/
void validate()
{
	// Search for existing uid.
	uint16_t pos;
	for(pos = 0; pos < 20; pos++)
	{
		// If exists, no Route Request Packet was send. 
		// Send it now!
		if(0 != list_cache_uid[pos] && list_cache_try[pos] < 4)
		{
			list_cache_try[pos]++;
			SendRouteRequest(list_cache_dst[pos], 1);
			if(DEBUG == 1)
				printf("Resend Packet with uid %u.\r\n", list_cache_uid[pos]);
		}
		// Packet already resend. Discard it!
		else
		{
			list_cache_uid[pos] = 0;
		}
	}
	// Validate every 5 seconds.
	Timers_add(5*1024, validate, 0xffff);
}

/**
 * @brief Initializing Funktion
 */
void Process_init()
{
	System_registerCallback( C_RADIO, ( fp_vp )Process_radioHandler );
	System_registerCallback( C_SENSOR, ( fp_vp )Process_sensorHandler );
	validate(); // Start Garbage Collector
}

/** 
* @brief Method: Route Request Packet
*
* Create, modify and send a route request packet.
* @param dst: Adress of destination node.
* @param reroute: (0) Create a packet and send.
*		  (1) Modify a packet and send. 
*/
void SendRouteRequest(uint16_t dst, uint16_t reroute)
{
	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);

	npsargs.netheader.to = ADRCOMP_WBROADCAST;		//  Send packet as personal broadcast.
	npsargs.netheader.type = REQUEST_PACKET;		//  Send packet as Route Request Packet.
	npsargs.netheader.flags = 0;

	// Packet does not exist, create it.
	if(reroute == 0)
	{
		// Create first sequence number.
		if(seq_no == 1)
		{
			seq_no = seq_no + (npsargs.netheader.from * 100);
		}

		// Deny id's over node_id *100, to eliminate duplicate id's from other nodes
		if(seq_no == ((npsargs.netheader.from * 100) + 100))
		{
			seq_no = (npsargs.netheader.from * 100) + 1;
		}

		uint16_t pos = 0;				// Current position in cache.
		bool free_pos = false;				// Cache is not full.

		// Save routing information.
		for(pos = 0; pos < 20 && free_pos == false; pos++)
		{
			if(DEBUG == 1)
				printf("Cache [%u]: %u\r\n", pos, list_cache_uid[pos]);

			// Search for free position in cache.
			if(list_cache_uid[pos] == 0)
			{
				list_cache_uid[pos] = seq_no;	
				list_cache_data[pos] = seq_no;
				list_cache_dst[pos] = dst;
				list_cache_try[pos] = 0;
				free_pos = true;
			}
			
		}

		// Cache is full.
		if(free_pos == false)
		{
			//printf("Packet canceled, cache is full!\r\n");
			return false;
		}
		

		list_id[seq_counter % 20] = seq_no;			// Insert sequence number in local uid list.
				
		route.source = npsargs.netheader.from;			// Set source adress.
		route.dest = dst;					// Set destination adress.
		route.hops[1] = npsargs.netheader.from;		// Set current node as first hop in packet header.
		route.hops[0] = 2;					// Increase index.
		
		route.id = seq_no++;					// Increase sequence number. (uid)
		
		seq_counter =  (seq_counter + 1) % 20;			// Increase seq_counter for next position in local uid list on node. 

		if(DEBUG == 1)
			printf("Sending packet with sequence number %u from source %u.\r\n", route.id, route.source);
	}
	// Packet exists, only modify it
	else
	{
		if(DEBUG == 1)
			printf("Route Request Packet (id %u) passed node %u to destination %u.\r\n", route.source, npsargs.netheader.from, route.dest);
		list_id[seq_counter % 20] = route.id;					// Insert sequence number in local uid list.
		route.hops[route.hops[0]] = npsargs.netheader.from;			// Set current node as next hop in packet header.
		++route.hops[0];							// Increase index.
		seq_counter =  (seq_counter + 1) % 20;					// Increase seq_counter for next position in local uid list on node. 
	}

	npsargs.payload[0].buffer = &route;				// Put request packet in buffer.
	npsargs.payload[0].size = sizeof(route);			// Determine size of packet.

	Net_send(&npsargs);						// Send packet.
}

/**
* @brief Method: Route Reply Packet
*
* Create, modify and send a route reply packet.
*
* If destination is reached, create and send the
* data packet.
*
*/
void SendRouteReply(struct netpacket_handler_args* args_copy)
{
	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);

	// Route Request Packet has reached destination. Sending Route Reply Packet.
	if(npsargs.netheader.from == route.dest)
	{
		if(DEBUG == 1)
			printf("Route Request Packet (id %u) has reached it's destination (id %u)!\r\n", route.source, route.dest);
		route.hops[route.hops[0]] = route.dest;
		route.hops[route.hops[0]+1] = 0;									// Sentinal for end position.
	}
	// Route Reply Packet has reached destination. Sending Data Packet.
	if(route.hops[0] == 1)
	{
		// Send data.
		if(DEBUG == 1)
			printf("Route Reply Packet (id %u) reached source (id %u).\r\nSending data packet.\r\n", route.dest, route.source);
		SendRouteData(args_copy, 1);
		
		// Check, if pending packets with the same uid exists in cache.
		// Send it.
		
		// Search for existing uid.
		uint16_t pos;
		for(pos = 0; pos < 20; pos++)
		{
			// Search packets withe same destination. 
			if(0 != list_cache_uid[pos] && list_cache_try[pos] < 4)
			{
				list_cache_try[pos]++;
				
				npsargs.netheader.to = list_cache_dst[pos];						//  Send packet with destination.
				npsargs.netheader.type = DATA_PACKET;							//  Send packet as Data Packet.
				npsargs.netheader.flags = 0;

				data.data = list_cache_data[pos];
				npsargs.payload[0].buffer = &data;							// Put Data Packet in buffer.
				npsargs.payload[0].size = sizeof(data);							// Determine size.
				if(DEBUG == 1)
					printf("Sending pending packets with same destination.\r\n"); 
				Net_send(&npsargs);									// Send packet.
			}

			// Packet already resend. Discard it!
			else
			{
				list_cache_uid[pos] = 0;
			}
		}
		
		return true;
	}
	// Route Reply Packet is on the way ...
	else
	{
		if(DEBUG == 1)
			printf("Route Reply Packet (id %u) passed node (id %u).\r\n", route.dest, npsargs.netheader.from);
		route.hops[0]--;						// Set next hop to forward packet to destination.
	}

	npsargs.netheader.to = route.hops[route.hops[0]];			//  Send packet as personal broadcast.
	npsargs.netheader.type = REPLY_PACKET;					//  Send packet as Route Reply Packet.
	npsargs.netheader.flags = 0;

	npsargs.payload[0].buffer = &route;					// Put request packet in buffer.
	npsargs.payload[0].size = sizeof(route);				// Determine size of packet.

	Net_send(&npsargs);						        // Send packet.
}

/**
* @brief Method: Data Packet
*
* Create, modify and send a data packet.
* If destination is reached, show payload data.
*
* @param args_copy copied data packet.
* @param CP Create packet if 1, else modify.	
*/
void SendRouteData(struct netpacket_handler_args* args_copy, int CP)
{
	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);
	
	// Create packet ...
	if(CP == 1)
	{
		// Search for correct data in local data cache.
		uint8_t counter = 0;
		while(list_cache_uid[counter] != route.id)
		{
			counter++;
		}

		// Fill Route Data Packet with data.
		data.data = list_cache_data[counter];

		// Delete data from local data cache.
		list_cache_uid[counter] = 0;

		// Set Next hop to forward packet.
		int k;
		for(k = 0; k < 11; k++)
		{
			data.hops[k] = route.hops[k];
		}

		++data.hops[0];
	}
	// ... else modify.
	else
	{
		// Set next hop to forward packet.
		++data.hops[0];
		// Data Packet reached destination.
		if(data.hops[data.hops[0]] == 0)
		{
			rec_counter++;
			if(DEBUG != 2)
				printf("Incoming data: %u.\r\n", data.data);
			if(DEBUG == 2)
				printf("RP: %u\r\n", rec_counter);
			return true;
		}
	}

	npsargs.netheader.to = data.hops[data.hops[0]];		//  Send packet with hop list.
	npsargs.netheader.type = DATA_PACKET;				//  Send packet as Data Packet.
	npsargs.netheader.flags = 0;

	npsargs.payload[0].buffer = &data;				// Put Data Packet in buffer.
	npsargs.payload[0].size = sizeof(data);				// Determine size.
	
	Net_send(&npsargs);						// Send packet.
}

/**
* @brief Send periodic packets every 3 seconds.
*/
void pTimer()
{
	SendRouteRequest(param, 0);
	if(DEBUG == 2)
	{
		if(req_counter == 100)
			printf("DONE\r\n");
		if(req_counter < 100)
		{
			Timers_add(1*1024, pTimer, param);
			req_counter++;
		}
	}
	else
		Timers_add(3*1024, pTimer, param);
}

/**
 * @brief Send TCP data  packets every second.
 * After that close connection to server. (FIN)
 * @param val sequence number of the packet.
 */
void TCPData(uint16_t val)
{
	if(DEBUG == 1)
	{
		printf("Left packets: %u\r\n", arg_packet_no);
	}
	int i;
	int empty2 = 0;
	bool full = false;
	for(i = 0; i < max_packets; i++)
	{
		if(list_seq_no_c[i] != 0)
		{
			empty2++;
		}
		if(empty2 == tcp_out.cw)
		{
			full = true;
		}	
	}
	if(full == false)
	{
		if(fin_send == false)
		{
			// Only send packet if we allowed to send
			if(arg_packet_no > 0)
			{
				tcp_out.ack_no = ++tcp_ack_no;
				if(DEBUG == 1)
				{
					printf("TCPData ...\r\ntcp_out.ack_no: %u\r\n", tcp_out.ack_no);
				}
				tcp_out.seq_no = ++tcp_seq_no;
				if(DEBUG == 1)
				{		
					printf("tcp_out.seq_no: %u\r\n", tcp_out.seq_no);
				}
				tcp_out.sync = false;
				tcp_out.ack = true;
				tcp_out.fin = false;
		
				tcp_seq_no = tcp_out.seq_no;
				tcp_ack_no = tcp_out.ack_no;
		
				if(DEBUG == 1)
				{
					printf("Client TCP\r\n");
					printf("seq_no: %u\r\nack_no: %u\r\nsync: %u\r\nack: %u\r\nfin: %u\r\n", tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
				}
		
				ModifyTCP(tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
				// Decrement packet count. So we only sends the amount of packets, we specified. 
				arg_packet_no--;
				// Every second send a packet.
				val = tcp_out.seq_no;
				
				if(c_window_counter < c_window)
				{
					Timers_add(4*1024, TCPData, tcp_seq_no);
					// Timeout to resend packet.
					Timers_add(10*1024, Resend, val);
					c_window_counter++;
				}
				
			}
			// Send the fin packet
			else
			{
				int i;
				bool empty = true;
				for(i = 0; i < max_packets; i++)
				{
					if(list_seq_no_c[i] != 0)
						empty = false;
				}
				
				if(DEBUG == 1)
				{
					printf("Empty: %u, Last Position: %u, Content: %u\r\n", empty, (counter_typed)-1, list_payload_c[(counter_typed)-1]);
				}
				
				 // FIN Packet bauen
				if(empty == true && list_payload_c[(counter_typed)-2] != 0)
				{
					tcp_out.ack_no = ++tcp_in.seq_no;
					tcp_out.seq_no = ++tcp_seq_no;
					tcp_out.sync = false;
					tcp_out.ack = false;
					tcp_out.fin = true;
		
					tcp_seq_no = tcp_out.seq_no;
					tcp_ack_no = tcp_out.ack_no;
		
					ModifyTCP(tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
		
					if(DEBUG == 1)
					{
						printf("Cached data: ");
						int i=0;
						while(i < list_pos_c)
						{
							printf("%u ", list_payload_c[i]);
							printf("%u ", list_seq_no_c[i]);
							i++;
						}
						printf("\r\n");
					}
					
					if(DEBUG == 0)
					{
						printf("Data transmission finished!\r\n");
						printf("Cached data: ");
						int i=0;
						while(i < list_pos_c)
						{
							printf("%u ", list_seq_no_c[i]);
							printf("%u ", list_payload_c[i]);
							i++;
						}
						printf("\r\n");
						printf("Stats: ");
						printf("typed packets: %u, send packets: %u, received packets: %u, packets resend: %u \r\n", counter_typed, counter_send, counter_received, counter_resend);
					}
		
					// Reset nach Verbindungsabbau
					last_ack = 0;
					list_pos_c = 0;
					tcp_seq_no = 0;
					tcp_ack_no = 0;
		            server = true;
		            c_window_counter = 1;
		            c_window = 1;
		            received_ack = 0;		
					counter_received = 0;		
					counter_send = 0;		
					counter_resend	= 0;		
					counter_data_t	= 0;
		            // FIN state
		            fin_send = true;
		            Timers_add(30*1024, fin_send_reset, 0xffdd);
				}
				else 
				{
					if(DEBUG == 1)
					{
						printf("Not finished !\r\n");
					}
					return true;
				}
			}
		}
		else
		// Discard packets that arrived after a fin packets (packets send by left resend timers)
		{
			if(DEBUG == 1)
			{
				printf("DISCARDING CACHED PACKET \r\n");
				printf("seq_no: %u\r\nack_no: %u\r\nsync: %u\r\nack: %u\r\nfin: %u\r\n", tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin);
			}
			return true;
		}
	}
	else
	{
		Timers_add(10*1024, Resend, val);	
	}	
}

/**
* @brief Only handle the first data packet (special case)
*/
void OnePacket()
{
	TCPData(0);
}

/**
 * @brief Fill packet payload with some stuff
 * 
 */
 void fill()
 {
 	int b;
	for(b = 0; b < 300; b++)
	{
		big_payload[b] = b;
	}	
 }
 
/**
* @brief Modify a TCP packet.
* Fill payload, modify flags, numbers and insert payload into the cache. After that send the packet!
* @param seq_no the sequenznumber of the packet
* @param ack_no the acknowledgenumber of te packet
* @param sync SYNC Flag
* @param ack ACK Flag
* @param fin FIN Flag
*/
void ModifyTCP(uint16_t seq_no, uint16_t ack_no, bool sync, bool ack, bool fin)
{
	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);

	tcp_out.seq_no = seq_no;
	tcp_out.ack_no = ack_no;
	tcp_out.sync = sync;
	tcp_out.ack = ack;
	tcp_out.fin = fin;
	// Set local congestion window size to new calculated size.
	tcp_out.cw = c_window;

	// Handle the Data packet
	if(tcp_out.sync == false && tcp_out.fin == false && tcp_out.ack == true && server == false)
	{
		bool resend = false;
		bool allreadysend = false;
		int i;
		for(i = 0; i < list_pos_c; i++)
		{
			if(list_seq_no_c[i] == tcp_out.seq_no)
				resend = true;
			if(list_payload_c[i] == tcp_out.seq_no)
				allreadysend = true;
		}

		// Insert Data into the cache of the client
		if(resend == false && allreadysend == false)
		{
			if(bigpacket==0)
			{
				tcp_out.payload = seq_no;
			}
			else
			{
				tcp_out.payload = big_payload;
			}
			// Caching data
			list_seq_no_c[list_pos_c] = tcp_out.seq_no;
			list_payload_c[list_pos_c] = tcp_out.payload;

			int i;
			if(DEBUG == 1)
			{
				printf("Client:\r\n");
			}
			
			for(i = 0; i < max_packets; i++)
			{
				if(DEBUG == 1)
				{
					if(list_seq_no_c[i] != 0)
						printf("Cache[%u]=%u %u\r\n", i, list_seq_no_c[i], list_payload_c[i]);
				}
			}
			list_pos_c++;
		}
		else
		{
			// Get the cached payload out of the cache.
			if(allreadysend == true && resend == true)
			{
				tcp_out.payload = list_payload_c[seq_no-2];
				if(DEBUG == 1)
				{
					printf("Fill payload to resend data.\r\n");
					printf("PAYLOAD: %u\r\nSEQ:%u\r\n", tcp_out.payload, seq_no-2);
				}
			}
			else
			{
				return true;
			}
		}
	}
	// No payload 
	else
		tcp_out.payload = NULL;

	if(DEBUG == 1)
	{
		printf("SENDING TCP PACKET\r\n");
		printf("seq_no: %u\r\nack_no: %u\r\nsync: %u\r\nack: %u\r\nfin: %u\r\ncw: %u\r\npayload: %u\r\n", tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin, tcp_out.cw, tcp_out.payload);
	}

	npsargs.netheader.to = tcp_adr;		
	npsargs.netheader.type = TCP_PACKET;					// Send packet as TCP Packet.
	npsargs.netheader.flags = 0;

	npsargs.payload[0].buffer = &tcp_out;						// Put TCP packet in buffer.
	npsargs.payload[0].size = sizeof(tcp_out);					// Determine size of packet.

	counter_send++;
	Net_send(&npsargs);
	// Only one packet send, call function to create a fin packet
	if (list_pos_c == 1)
	{
		tcp_ack_no = tcp_out.ack_no;
		tcp_seq_no = tcp_out.seq_no;
		Timers_add(2*1024, OnePacket, 0xffff);
	}
}

/**
* @brief Send a TCP Packet to a specified destination.
* @param dst the destination of the packet
*/
void SendTCP(uint16_t dst)
{
    netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);

	tcp_out.seq_no = ++tcp_seq_no;
	tcp_out.ack_no = 0;
	tcp_out.sync = true;
	tcp_out.ack = false;
	tcp_out.fin = false;

	if(DEBUG == 1)
	{
		printf("Create a TCP Packet ...\r\n");
		printf("seq_no: %u\r\nack_no: %u\r\nsync: %u\r\nack: %u\r\nfin: %u\r\npayload: %u\r\n", tcp_out.seq_no, tcp_out.ack_no, tcp_out.sync, tcp_out.ack, tcp_out.fin, tcp_out.payload);
	}
	
	if(DEBUG == 0)
	{
		printf("Start transmission ...\r\n");
	}
	counter_send++;
	npsargs.netheader.to = dst;				//  Send packet to destination.
	npsargs.netheader.type = TCP_PACKET;	//  Send packet as TCP Packet.
	npsargs.netheader.flags = 0;

	npsargs.payload[0].buffer = &tcp_out;		// Put TCP packet in buffer.
	npsargs.payload[0].size = sizeof(tcp_out);	// Determine size of packet.

	Net_send(&npsargs);
}
/* ----------------------------------------------------------------------------------------------------- */
/* ---------------------------------------- COMMANDS --------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------- */

/**
* @brief Command: bigpacket
*
* Set bigpacket mode on/off.
* @param	(0) OFF
*			(1) ON			
*/
COMMAND (bigpacket, CMDFLAG_SERIAL, cmdargs) 
{
	bigpacket = String_parseUint16( cmdargs->args, NULL );
}


/**
* @brief Command: debug
*
* Set debug mode on/off.
* @param	(0) OFF
*			(1) ON			
*/
COMMAND (debug, CMDFLAG_SERIAL, cmdargs) 
{
	DEBUG = String_parseUint16( cmdargs->args, NULL );
}

/**
* @brief Command: SendTCP
*
* Send a TCP Packet.
* @param Destination adress.
*/
COMMAND (SendTCP, CMDFLAG_SERIAL, cmdargs) 
{
	uint16_t count = Cmd_SplitArgs( cmdargs->args );
	uint16_t* dst = (uint16_t*) cmdargs->args;

	// Save server adress from command prompt.
	tcp_adr = String_parseUint16(dst, NULL );

	// This node is the client.
	server = false;
	
	// Reset FIN state.
	fin_send = false;
	// Reset counters.
	received_ack = 0;		
	counter_received = 0;		
	counter_send = 0;		
	counter_resend	= 0;		
	counter_data_t	= 0;

	// User wants to send more than one packet to server.
	if(count > 1)
	{
		// Save number of packets to send.
		arg_packet_no = String_parseUint16( Cmd_GetNextArg( cmdargs->args ), NULL ); 
	}

	counter_typed = arg_packet_no;
	// Check if we transmit big packets.
	if (bigpacket==true)
	{
		printf("bigpacket mode enabled\r\n");
		fill();
	}
	// Actually user can only send max_packets packets to server, because of implementation details. 
	if(arg_packet_no <= max_packets)
	{
		// Reset earlier received packets.
		last_ack = 0;
                c_window = 1;
                list_pos_c = 0;
                c_window_counter = 1;
		SendTCP(String_parseUint16(dst, NULL ));
	}
	else
	{
	printf("Sorry, more than %u packets are not allowed!\r\n", max_packets);
	}
}

/**
* @brief Command: Packet
*
* Send a packet with DSR.
* @param Destination adress.
*/
COMMAND (Packet, CMDFLAG_SERIAL, cmdargs) 
{
	SendRouteRequest(String_parseUint16( cmdargs->args, NULL ), 0);
}

/**
* @brief Command: RouteTest
* 
* Send periodical packet with DSR.
* @param Destination adress. (1 - 9)
*/
COMMAND (RouteTest, CMDFLAG_SERIAL, cmdargs) 
{
	param = String_parseUint16( cmdargs->args, NULL );
	
	// Stop timer (@param 0)
	if(param == 0)
	{
		Timers_remove(pTimer, oldParam);
		oldParam = 0;
	}
	else
	{
		// First Timer
		if(oldParam == 0)
		{
			oldParam = param;
			Timers_add(3*1024, pTimer, param);
		}
		// Override Timer with new destination.
		else
		{
			Timers_remove(pTimer, oldParam);
			Timers_add(3*1024, pTimer, param);
			oldParam = param;
		}
	}
}

/**
* @brief Promiscious mode
* 
* Enabled  (1)
* Disabled (0)
*/
COMMAND(Prom, CMDFLAG_SERIAL, cmdargs)
{
	uint16_t p = String_parseUint16( cmdargs->args, NULL );
	if(p == 0 || p == 1)
		Promiscous = p;
	else
		printf("Error in parameter!\n");
}

/**
* @brief Ping request
*
* Send a ping request to another node
* Example: ping 24
*/
COMMAND( ping, 0, cmdargs )
{
	uint16_t node;

	node = String_parseUint16( cmdargs->args, NULL );
	if ( node > 0 ) {
		Net_sendPing((netaddr_t)node);
	}
}

/**
* @brief Echo Arguments
*
* Testcase 1: long command name
* Testcase 2: argument splitting
*/
COMMAND( echoargs, CMDFLAG_SERIAL, cmdargs )
{
	uint16_t i;
	uint16_t count = Cmd_SplitArgs( cmdargs->args );
	char* arg = (char*) cmdargs->args;

	for ( i = 1; i <= count; i++ ) {
		printf("%u: %*s\n", i, strlen(arg), arg);
		arg = ( char* )Cmd_GetNextArg( arg );
	}
	snprintf(cmdargs->response_args, cmdargs->response_bufsize, "%u args found", count );
}

// Software header.
APP(
    APPH_STATIC_APP,	// Application Type
    __APP_NAME,			// Application Name (8 chars)
    1, 0				// Aplication Version (2 bytes)
);
